/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// This code in this file converts multi-byte encodings from UTF-8 into HTML
// character entities of the form &#####;, where ##### is the number
// that the multi-byte sequence encodes.
//
// This is a work-around for a problem that I haven't been able to
// find the right solution to.  The problem is that the Dict extension
// is unable to tell Mozilla/Firefox that the content it is displaying
// is UTF-8.  Mozilla/Firefox then mis-interprets it as an 8-bit coding
// from the ISO-8859-1 (Latin-1) character set, causing garbage to be 
// displayed for words with non-ASCII characters.


function dictConvertUtf8ToOtherCodingInternal(utf8String, convertToCharEntity) {

    var encodedString = '';
    var octetsRemainingInCharacterCode = 0;
    var characterCode = 0;

    for (var i = 0; i < utf8String.length; i++) {

        var inputCharCode = utf8String.charCodeAt(i);

        if ((inputCharCode & 0x80) == 0x00) {
            // this can be encoded with ASCII-7
            encodedString += utf8String[i];

            // reset the multi-octet stream accumulator.
            // if it was not zero, then the input was bad.
            octetsRemainingInCharacterCode = 0;
        }
        else if ((inputCharCode & 0xC0) == 0x80) {
            // 10xxxxxx

            // this is a continuation of a multi-octet sequence
            characterCode = (characterCode << 6) | (inputCharCode & 0x3F);

            octetsRemainingInCharacterCode--;
            if (octetsRemainingInCharacterCode == 0) {
                // this is the final octet in the mutli-octet code

                if (convertToCharEntity) {
                    // replace the UTF-8 coding with an HTML character entity
                    encodedString += "&#" + characterCode + ";";
                }
                else {
                    // replace the UTF-8 coding with its Unicode value
                    encodedString += String.fromCharCode(characterCode);
                }
            }
            else if (octetsRemainingInCharacterCode < 0) {
                // bad input
                octetsRemainingInCharacterCode = 0;
            }
        }
        else if ((inputCharCode & 0xE0) == 0xC0) {
            // 110xxxxx 10xxxxxx
            characterCode = inputCharCode & 0x1F;
            octetsRemainingInCharacterCode = 1;
        }
        else if ((inputCharCode & 0xF0) == 0xE0) {
            // 1110xxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x0F;
            octetsRemainingInCharacterCode = 2;
        }
        else if ((inputCharCode & 0xF8) == 0xF0) {
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x07;
            octetsRemainingInCharacterCode = 3;
        }
        else if ((inputCharCode & 0xFC) == 0xF8) {
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x03;
            octetsRemainingInCharacterCode = 4;
        }
        else if ((inputCharCode & 0xFE) == 0xFC) {
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x01;
            octetsRemainingInCharacterCode = 5;
        }
        else {
            // Illegal UTF-8 code.  Drop it.
        }
    }

    if (octetsRemainingInCharacterCode != 0) {

        // The UTF-8 seems to have been cut short.
        // kassiopeia.juls.savba.sk sometimes does this.
        // rather than reject the server, we are tolerant of bad input
        if (convertToCharEntity) {
            // replace the UTF-8 coding with an HTML character entity
            encodedString += "&#" + characterCode + ";";
        }
        else {
            // replace the UTF-8 coding with its Unicode value
            encodedString += String.fromCharCode(characterCode);
        }
    }

    return encodedString;
}

//
// dictUtf8ToOtherCoding() converts an octet stream that is 
// encoded in UTF-8 into an ASCII-7 encoding of the same 
// stream such that all character > 127 are encoded using
// HTML character entities &#xxx;.
//
// This is necessary when using international DICT servers,
// because I cannot figure out another way to tell Mozilla
// to interpret the string as UTF-8.
//
//   UCS-4 range (hex.)    UTF-8 octet sequence (binary)
//   0000 0000-0000 007F   0xxxxxxx
//   0000 0080-0000 07FF   110xxxxx 10xxxxxx
//   0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
//   0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//
// Parameters:
//   utf8String : string -
//      the UTF-8 string to encode
//
//   convertToCharEntity : bool
//      if true, the return value will be encoded as HTML,
//      with all non-ASCII-7 values represented as
//      character entities (of the form "&#<n>;")
//
//      if false, the return value will be encoded as UCS-4,
//      which are normal JavaScript strings.
//
function dictUtf8ToOtherCoding(utf8String, convertToCharEntity) {

    // Before we do any convertion, let's see if there are
    // any character codes outside of the ASCII-7 range.
    // If not, then the string doesn't need to be converted.

    if (!utf8String.match(/[\x80-\xFF]/)) {
        return utf8String;
    }

    // Note that we could just call utf8ToCharacterEntitiesInternal()
    // on the entire string, but this routine is very slow.
    // Instead, it is faster to use a regular expression
    // to separate the octet-sequences that must be replaced
    // while preserving long runs of characters that don't to escaping.

    // This makes the routine run in O(n^2) of the characters that
    // must be replaced, instead of O(n^2) of all characters.
    // When the text is mostly ASCII-7, this is a significant savings.

    var encodedString = '';
    
    var matchExp = /([\x00-\x7F]*)([\x80-\xFF]+)([\x00-\x7F]*)/g;
    while ((result = matchExp.exec(utf8String)) != null) {

        var singleOctetCodesBefore = result[1];
        var multiOctetCodes        = result[2];
        var singleOctetCodesAfter  = result[3];
        
        encodedString += 
            singleOctetCodesBefore + 
            dictConvertUtf8ToOtherCodingInternal(multiOctetCodes, convertToCharEntity) +
            singleOctetCodesAfter;
    }

    return encodedString;
}


/////////////////////////////////////////////////////////////////////
// Public Functions


//
// dictUtf8ToCharacterEntities() converts an octet stream that is 
// encoded in UTF-8 into an ASCII-7 encoding of the same 
// stream such that all character > 127 are encoded using
// HTML character entities &#xxx;.
//
function dictUtf8ToCharacterEntities(utf8String) {
    return dictUtf8ToOtherCoding(utf8String, true);
}

//
// dictUtf8ToUnicode() converts an octet stream to unicode.
//
function dictUtf8ToUcs4(utf8String) {
    return dictUtf8ToOtherCoding(utf8String, false);
}


// dictUcs4ToUtf8() returns a UTF-8 representation of ucs4String
function dictUcs4ToUtf8(ucs4String) {

    var utf8String = '';
    for (var i = 0; i < ucs4String.length; i++) {
        var charCode = ucs4String.charCodeAt(i);

        if (0x7FFFFFFF < charCode) {
            // This byte cannot be encoded in UTF-8.
            // Skip it
            // NOTE: this should never happen in practice
            // because JavaScript uses UCS-4.
        }
        else if (0x3FFFFFF < charCode) {
            // six-byte UTF-8 sequence
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xFC | ((charCode >> 30) & 0x01),
                0x80 | ((charCode >> 24) & 0x3F),
                0x80 | ((charCode >> 18) & 0x3F),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x1FFFFF < charCode) {
            // five-byte UTF-8 sequence
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xF8 | ((charCode >> 24) & 0x03),
                0x80 | ((charCode >> 18) & 0x3F),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0xFFFF < charCode) {
            // four-byte UTF-8 sequence
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xF0 | ((charCode >> 18) & 0x07),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x7FF < charCode) {
            // three-byte UTF-8 sequence
            // 1110xxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xE0 | ((charCode >> 12) & 0x0F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x7F < charCode) {
            // two-byte UTF-8 sequence
            // 110xxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xC0 | ((charCode >>  6) & 0x1F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else {
            // one-byte UTF-8 sequence
            utf8String += ucs4String[i];
        }
    }

    return utf8String;
}
